# NOTE: implementation is based on https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-dnsp/445c7843-e4a1-4222-8c0f-630c230a4c80

from __future__ import annotations

import sys
from typing import ClassVar, Generic, List, Optional, TypeVar

from msldap.commons.utils import timestamp2datetime
from msldap.wintypes.dnsp.structures.misc import DnsAddrArray, Ip4Array

# NOTE: typing.Type is deprecated since version 3.9
if sys.version_info < (3, 9):
    from typing import Type
else:
    Type = type


# NOTE: to make typing happy with auto-implementation of from_bytes
class _Unserializable:
    @classmethod
    def from_bytes(cls, data: bytes, byteorder: str) -> _Unserializable:
        ...


# NOTE: by default, consider BaseType unserializable to make possible the generation of from_bytes method
BaseType = TypeVar("BaseType", bound=_Unserializable)


class BaseDnsProperty(Generic[BaseType]):
    _id: ClassVar[int] = None
    _name: ClassVar[str] = None

    _base_type: ClassVar[Type[BaseType]] = None
    _default: ClassVar[BaseType] = None

    _implementations: ClassVar[List[Type[BaseDnsProperty]]] = []

    def __init__(self, value: Optional[BaseType] = None) -> None:
        self._value = value if value is not None else self._default

    def __str__(self) -> str:
        return f"{self._name}: {self.value}"

    @property
    def value(self) -> str:
        return str(self._value)

    @classmethod
    def from_bytes(cls, data: bytes) -> BaseDnsProperty:
        return cls(cls._base_type.from_bytes(data, "little"))

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)

        cls._base_type = cls.__orig_bases__[0].__args__[0]

        if (cls._id is not None) and (cls._name is not None):
            cls._implementations.append(cls)


class DnsEnumProperty(BaseDnsProperty[int]):
    _types: ClassVar[dict[int, str]] = None

    @property
    def value(self) -> str:
        return self._types.get(self._value, "Unknown")


class DnsFlagsProperty(BaseDnsProperty[int]):
    _flags: ClassVar[dict[int, str]] = None

    @property
    def value(self) -> str:
        return " | ".join(
            value for (key, value) in self._flags.items() if (self._value & key)
        )


class DnsHoursIntervalProperty(BaseDnsProperty[int]):
    @property
    def value(self) -> str:
        if self._value == 0:
            return "0 hours"

        result = ""

        days = self._value // 24
        if days != 0:
            result += f"{days} days "

        hours = self._value % 24
        if hours != 0:
            result += f"{hours} hours "

        return result.rstrip()


class DnsTimestampProperty(BaseDnsProperty[int]):
    @property
    def value(self) -> str:
        if self._value == 0:
            return "Never"

        return timestamp2datetime(self._value).isoformat()


class DnsNullTerminatedStringProperty(BaseDnsProperty[str]):
    @classmethod
    def from_bytes(cls, data: bytes) -> DnsNullTerminatedStringProperty:
        return cls(data.decode("utf-8").rstrip("\x00"))


# NOTE: https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-dnsp/27e138a7-110c-44a4-afcb-b95f35f00306
class DnsZoneType(DnsEnumProperty):
    _id = 0x01
    _name = "Zone Type"

    _default = 1

    _types = {
        0: "Cache",
        1: "Primary",
        2: "Secondary",
        3: "Stub",
        4: "Forwarder",
        5: "Secondary Cache",
    }


# NOTE: https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-dnsp/d4b84209-f00c-478f-80d7-8dd0f1633d9e
class DnsZoneAllowUpdate(DnsEnumProperty):
    _id = 0x02
    _name = "Dynamic Updates"

    _default = -1

    _types = {
        0: "Not allowed",
        1: "All updates allowed",
        2: "Only secure updates allowed",
    }


class DnsZoneSecureTime(DnsTimestampProperty):
    _id = 0x08
    _name = "Time Zone Secured"

    _default = 0


class DnsZoneNoRefreshInterval(DnsHoursIntervalProperty):
    _id = 0x10
    _name = "Zone No-Refresh Interval"

    _default = 168


class DnsZoneRefreshInterval(DnsHoursIntervalProperty):
    _id = 0x20
    _name = "Zone Refresh Interval"

    _default = 168


class DnsZoneAgingState(BaseDnsProperty[bool]):
    _id = 0x40
    _name = "Aging Enabled"

    _default = False


class DnsZoneScavengingServers(BaseDnsProperty[Ip4Array]):
    _id = 0x11
    _name = "DNS Servers performing Scavenging"

    _default = Ip4Array()


class DnsZoneAgingEnabledTime(DnsHoursIntervalProperty):
    _id = 0x12
    _name = "Time before the next Scavenging Cycle"

    _default = 0


class DnsZoneDeletedFromHostname(DnsNullTerminatedStringProperty):
    _id = 0x80
    _name = "Name of Server that deleted the Zone"

    _default = "Unknown"


class DnsZoneMasterServers(BaseDnsProperty[Ip4Array]):
    _id = 0x81
    _name = "DNS Servers performing Zone Transfers"

    _default = Ip4Array()


class DnsZoneAutoNsServers(BaseDnsProperty[Ip4Array]):
    _id = 0x82
    _name = "DNS Servers which may autocreate a Delegation"

    _default = Ip4Array()


# NOTE: https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-dnsp/4ec7bdf7-1807-4179-96af-ce1c1cd448b7
class DnsZoneDcPromoConvert(DnsEnumProperty):
    _id = 0x83
    _name = "State of Conversion"

    _default = 0

    _types = {
        0: "None",
        1: "To be moved to DNS Domain partition",
        2: "To be moved to DNS Forest partition",
    }


class DnsZoneScavengingServersDa(BaseDnsProperty[DnsAddrArray]):
    _id = 0x90
    _name = "DNS Servers performing Scavenging"

    _default = DnsAddrArray()


class DnsZoneMasterServersDa(BaseDnsProperty[DnsAddrArray]):
    _id = 0x91
    _name = "DNS Servers performing Zone Transfers"

    _default = DnsAddrArray()


class DnsZoneAutoNsServersDa(BaseDnsProperty[DnsAddrArray]):
    _id = 0x92
    _name = "DNS Servers which may autocreate a Delegation"

    _default = DnsAddrArray()


# NOTE: https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-dnsp/f448341f-512d-414a-aaa3-e303d592fcd2
class DnsZoneNodeDbFlags(DnsFlagsProperty):
    _id = 0x100
    _name = "DB Flags"

    _default = 0

    _flags = {
        0x80000000: "DNS_RPC_FLAG_CACHE_DATA",
        0x40000000: "DNS_RPC_FLAG_ZONE_ROOT",
        0x20000000: "DNS_RPC_FLAG_AUTH_ZONE_ROOT",
        0x10000000: "DNS_RPC_FLAG_ZONE_DELEGATION",
        0x08000000: "DNS_RPC_FLAG_RECORD_DEFAULT_TTL",
        0x04000000: "DNS_RPC_FLAG_RECORD_TTL_CHANGE",
        0x02000000: "DNS_RPC_FLAG_RECORD_CREATE_PTR",
        0x01000000: "DNS_RPC_FLAG_NODE_STICKY",
        0x00800000: "DNS_RPC_FLAG_NODE_COMPLETE",
        0x00010000: "DNS_RPC_FLAG_SUPPRESS_NOTIFY",
        0x00020000: "DNS_RPC_FLAG_AGING_ON",
        0x00040000: "DNS_RPC_FLAG_OPEN_ACL",
        0x00100000: "DNS_RPC_FLAG_RECORD_WIRE_FORMAT",
        0x00200000: "DNS_RPC_FLAG_SUPPRESS_RECORD_UPDATE_PTR",
    }


class DnsPropertyFactory:
    def __init__(self) -> None:
        self._factories = {
            subclass._id: subclass for subclass in BaseDnsProperty._implementations
        }

    def from_bytes(self, data: bytes) -> Optional[BaseDnsProperty]:
        length = int.from_bytes(data[:4], "little")
        id = int.from_bytes(data[16:20], "little")

        FactoryClass = self._factories.get(id)
        if FactoryClass is None:
            return None

        if length == 0:
            return FactoryClass()

        return FactoryClass.from_bytes(data[20 : 20 + length])




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import enum
import copy

from msldap.commons.target import MSLDAPTarget
from msldap.client import MSLDAPClient
from msldap.connection import MSLDAPClientConnection
from asyauth.common.credentials import UniCredential

class LDAPConnectionFactory:
	"""
	The URL describes both the connection target and the credentials. This class creates all necessary objects to set up the client.
	
	:param url: 
	:type url: str
	"""

	help_epilog = """
	MSLDAP URL Format: <protocol>+<auth>://<username>:<password>@<ip_or_host>:<port>/<tree>/?<param>=<value>
	<protocol> sets the ldap protocol following values supported:
		- ldap
		- ldaps
	<auth> can be omitted if plaintext authentication is to be performed (in that case it default to ntlm-password), otherwise:
		- ntlm-password
		- ntlm-nt
		- kerberos-password (dc option param must be used)
		- kerberos-rc4 / kerberos-nt (dc option param must be used)
		- kerberos-aes (dc option param must be used)
		- kerberos-keytab (dc option param must be used)
		- kerberos-ccache (dc option param must be used)
		- sspi-ntlm (windows only!)
		- sspi-kerberos (windows only!)
		- anonymous
		- plain
		- simple
		- sicily (same format as ntlm-nt but using the SICILY authentication)
	<tree>:
		OPTIONAL. Specifies the root tree of all queries
	<param> can be:
		- timeout : connction timeout in seconds
		- proxytype: currently only socks5 proxy is supported
		- proxyhost: Ip or hostname of the proxy server
		- proxyport: port of the proxy server
		- proxytimeout: timeout in secodns for the proxy connection
		- dc: the IP address of the domain controller, MUST be used for kerberos authentication
		- encrypt: enable encryption. Only for NTLM. DOESNT WORK WITH LDAPS
		- etype: Supported encryption types for Kerberos authentication. Multiple can be specified.
		- rate: LDAP paged search query rate limit. Will sleep for seconds between each new page. Default: 0 (no limit)
		- pagesize: LDAP paged search query size per page. Max: 1000. Default: 1000

	Examples:
	ldap://10.10.10.2 (anonymous bind)
	ldaps://test.corp (anonymous bind)
	ldap+sspi-ntlm://test.corp
	ldap+sspi-kerberos://test.corp
	ldap://TEST\\victim:<password>@10.10.10.2 (defaults to SASL GSSAPI NTLM)
	ldap+simple://TEST\\victim:<password>@10.10.10.2 (SASL SIMPLE auth)
	ldap+plain://TEST\\victim:<password>@10.10.10.2 (SASL SIMPLE auth)
	ldap+ntlm-password://TEST\\victim:<password>@10.10.10.2
	ldap+ntlm-nt://TEST\\victim:<nthash>@10.10.10.2
	ldap+kerberos-password://TEST\\victim:<password>@10.10.10.2
	ldap+kerberos-rc4://TEST\\victim:<rc4key>@10.10.10.2
	ldap+kerberos-aes://TEST\\victim:<aes>@10.10.10.2
	ldap://TEST\\victim:password@10.10.10.2/DC=test,DC=corp/
	ldap://TEST\\victim:password@10.10.10.2/DC=test,DC=corp/?timeout=99&proxytype=socks5&proxyhost=127.0.0.1&proxyport=1080&proxytimeout=44
	"""
	
	def __init__(self, credential:UniCredential = None, target:MSLDAPTarget = None ):
		self.credential = credential
		self.target = target

	@staticmethod
	def from_url(connection_url):
		target = MSLDAPTarget.from_url(connection_url)
		credential = UniCredential.from_url(connection_url)
		return LDAPConnectionFactory(credential, target)

	def get_credential(self) -> UniCredential:
		"""
		Creates a credential object
		
		:return: Credential object
		:rtype: :class:`UniCredential`
		"""
		return copy.deepcopy(self.credential)

	def get_target(self) -> MSLDAPTarget:
		"""
		Creates a target object
		
		:return: Target object
		:rtype: :class:`MSLDAPTarget`
		"""
		return copy.deepcopy(self.target)

	def get_client(self) -> MSLDAPClient:
		"""
		Creates a client that can be used to interface with the server
		
		:return: LDAP client
		:rtype: :class:`MSLDAPClient`
		"""
		cred = self.get_credential()
		target = self.get_target()
		return MSLDAPClient(target, cred)


	def get_connection(self) -> MSLDAPClientConnection:
		"""
		Creates a connection that can be used to interface with the server
		
		:return: LDAP connection
		:rtype: :class:`MSLDAPClientConnection`
		"""
		cred = self.get_credential()
		target = self.get_target()
		return MSLDAPClientConnection(target, cred)
	
	@staticmethod
	def from_ldapconnection(connection:MSLDAPClientConnection):
		"""Creates a new LDAPConnectionFactory object from an existing SMBConnection object"""
		"""This is useful when you have a connection object, but you need to create a new connection with the same credentials"""
		return LDAPConnectionFactory(copy.deepcopy(connection.credential), copy.deepcopy(connection.target))

		
	def __str__(self):
		t = '==== LDAPConnectionFactory ====\r\n'
		for k in self.__dict__:
			val = self.__dict__[k]
			if isinstance(val, enum.IntFlag):
				val = val
			elif isinstance(val, enum.Enum):
				val = val.name
			
			t += '%s: %s\r\n' % (k, str(val))
			
		return t

if __name__ == '__main__':
	url_tests = [
		'ldap://10.10.10.2',
		'ldap://10.10.10.2:9999',
		'ldap://test:password@10.10.10.2',
		'ldap://domain\\test@10.10.10.2',
		'ldap://domain\\test:password@10.10.10.2:9999',
		'ldap://domain\\test:AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA@10.10.10.2:9999',
		'ldaps+sspi-ntlm://10.10.10.2',
		'ldaps+sspi-kerberos://10.10.10.2',
		'ldaps+ntlm-password://domain\\test:password@10.10.10.2:9999',
		'ldaps+ntlm-nt://domain\\test:password@10.10.10.2:9999',
		'ldaps+kerberos-password://domain\\test:password@10.10.10.2:9999',
		'ldaps://10.10.10.2:9999',
		'ldaps://test:password@10.10.10.2',
		'ldaps://domain\\test@10.10.10.2',
		'ldaps://domain\\test:password@10.10.10.2:9999',
		'ldaps://DOMAIN\\test:password@10.10.10.2:9999/?proxytype=socks5&proxyserver=127.0.0.1',
		'ldaps://DOMAIN\\test:password@10.10.10.2:9999/?proxytype=socks5&proxyserver=127.0.0.1&proxyuser=admin&proxypass=alma',
		'ldaps://DOMAIN\\test:password@10.10.10.2:9999/?proxytype=multiplexor&proxyserver=127.0.0.1&proxyport=9999&proxyuser=admin&proxypass=alma',
		'ldaps://10.10.10.2',
		'ldaps://10.10.10.2:6666',
	]
	for url in url_tests:
		print('===========================================================================')
		print(url)
		try:
			dec = LDAPConnectionFactory.from_url(url)
			creds = dec.get_credential()
			target = dec.get_target()
		except Exception as e:
			import traceback
			traceback.print_exc()
			print('ERROR! Reason: %s' % e)
			input()
		else:
			print(str(creds))
			print(str(target))
			input()



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from msldap.ldap_objects.adinfo import MSADInfo, MSADInfo_ATTRS
from msldap.ldap_objects.aduser import MSADUser, MSADUser_ATTRS, MSADUser_TSV_ATTRS
from msldap.ldap_objects.adcomp import MSADMachine, MSADMachine_ATTRS, MSADMachine_TSV_ATTRS
from msldap.ldap_objects.adsec  import MSADSecurityInfo, MSADTokenGroup
from msldap.ldap_objects.common import MSLDAP_UAC
from msldap.ldap_objects.adgroup import MSADGroup, MSADGroup_ATTRS
from msldap.ldap_objects.adou import MSADOU, MSADOU_ATTRS
from msldap.ldap_objects.adgpo import MSADGPO, MSADGPO_ATTRS
from msldap.ldap_objects.adtrust import MSADDomainTrust, MSADDomainTrust_ATTRS
from msldap.ldap_objects.adschemaentry import MSADSCHEMAENTRY_ATTRS, MSADSchemaEntry
from msldap.ldap_objects.adca import MSADCA, MSADCA_ATTRS
from msldap.ldap_objects.adenrollmentservice import MSADEnrollmentService_ATTRS, MSADEnrollmentService
from msldap.ldap_objects.adcertificatetemplate import MSADCertificateTemplate, MSADCertificateTemplate_ATTRS
from msldap.ldap_objects.adgmsa import MSADGMSAUser, MSADGMSAUser_ATTRS
from msldap.ldap_objects.adcontainer import MSADContainer, MSADContainer_ATTRS


__all__ = [
    'MSADUser', 
    'MSADUser_ATTRS', 
    'MSADUser_TSV_ATTRS', 
    'MSADInfo',
    'MSADInfo_ATTRS',
    'MSLDAP_UAC',
    'MSADMachine', 
    'MSADMachine_ATTRS',
    'MSADMachine_TSV_ATTRS',
    'MSADSecurityInfo',
    'MSADTokenGroup',
    'MSADGroup',
    'MSADOU', 
    'MSADGPO',
    'MSADGPO_ATTRS',
    'MSADDomainTrust',
    'MSADDomainTrust_ATTRS',
    'MSADGroup_ATTRS',
    'MSADOU_ATTRS',
    'MSADSCHEMAENTRY_ATTRS',
    'MSADSchemaEntry',
    'MSADCA',
    'MSADCA_ATTRS',
    'MSADEnrollmentService_ATTRS',
    'MSADEnrollmentService',
    'MSADCertificateTemplate', 
    'MSADCertificateTemplate_ATTRS',
    'MSADGMSAUser', 
    'MSADGMSAUser_ATTRS',
    'MSADContainer',
    'MSADContainer_ATTRS',

]


import io
import enum
from asn1crypto.cms import ContentInfo

class EncryptedLAPSBlob:
	def __init__(self):
		self.update_timestamp: int
		self.flags: int
		self.blob: bytes
		self.asn1blob: ContentInfo
	
	def get_keyidentifier(self):
		sid = self.asn1blob.native['content']['recipient_infos'][0]['kekid']['other']['key_attr']['1']['0']['0']['1']
		print(sid)
		return LAPS_KEYIDENTIFIER.from_bytes(self.asn1blob.native['content']['recipient_infos'][0]['kekid']['key_identifier'])
	
	def from_bytes(data: bytes):
		return EncryptedLAPSBlob.from_buffer(io.BytesIO(data))
	
	def from_buffer(buff: io.BytesIO):
		blob = EncryptedLAPSBlob()
		blob.update_timestamp = int.from_bytes(buff.read(8), byteorder='little', signed=False)
		blob_length = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		blob.flags = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		blob.blob = buff.read(blob_length)
		blob.asn1blob = ContentInfo.load(blob.blob)
		return blob
	
	def __str__(self):
		t = ''
		for k in self.__dict__:
			t += '%s: %s\n' % (k, self.__dict__[k])
		return t

class KEYIDFLAGS(enum.IntFlag):
	DHPARAMS = 1
	UNKNOWN = 2

	
class LAPS_KEYIDENTIFIER:
	def __init__(self):
		self.version = None
		self.magic = None
		self.flags = None
		self.l0_index = None
		self.l1_index = None
		self.l2_index = None
		self.root_key_identifier = None
		self.unknown_length = None
		self.domain_length = None
		self.forest_length = None
		self.unknown = None
		self.domain = None
		self.forest = None
	
	@staticmethod
	def from_bytes(data:bytes):
		return LAPS_KEYIDENTIFIER.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		blob = LAPS_KEYIDENTIFIER()
		blob.version = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		blob.magic = buff.read(4)
		blob.flags = KEYIDFLAGS(int.from_bytes(buff.read(4), byteorder='little', signed=False))
		blob.l0_index = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		blob.l1_index = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		blob.l2_index = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		blob.root_key_identifier = buff.read(16)
		blob.unknown_length = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		blob.domain_length = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		blob.forest_length = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		blob.unknown = buff.read(blob.unknown_length)
		blob.domain = buff.read(blob.domain_length)
		blob.forest = buff.read(blob.forest_length)
		return blob
	
	def __str__(self):
		t = ''
		for k in self.__dict__:
			if isinstance(self.__dict__[k], bytes):
				t += '%s: %s\n' % (k, self.__dict__[k].hex())
			else:
				t += '%s: %s\n' % (k, self.__dict__[k])
		return t


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import enum
from datetime import datetime

def vn(x):
	"""
	value or none, returns none if x is an empty list
	"""
	if x == []:
		return None
	if isinstance(x, list):
		return '|'.join(x)
	if isinstance(x, datetime):
		return x.isoformat()
	return x

class MSLDAP_UAC(enum.IntFlag):
	SCRIPT = 0x00000001  #[ADS_UF_SCRIPT](https://msdn.microsoft.com/library/aa772300) 	The logon script is executed.
	ACCOUNTDISABLE = 0x00000002  #[ADS_UF_ACCOUNTDISABLE](https://msdn.microsoft.com/library/aa772300) 	The user account is disabled.
	HOMEDIR_REQUIRED = 0x00000008  #[ADS_UF_HOMEDIR_REQUIRED](https://msdn.microsoft.com/library/aa772300) 	The home directory is required.
	LOCKOUT = 0x00000010  #[ADS_UF_LOCKOUT](https://msdn.microsoft.com/library/aa772300) 	The account is currently locked out.
	PASSWD_NOTREQD = 0x00000020  #[ADS_UF_PASSWD_NOTREQD](https://msdn.microsoft.com/library/aa772300) 	No password is required.
	PASSWD_CANT_CHANGE = 0x00000040  #[ADS_UF_PASSWD_CANT_CHANGE](https://msdn.microsoft.com/library/aa772300) 	The user cannot change the password.
	ENCRYPTED_TEXT_PASSWORD_ALLOWED = 0x00000080  #[ADS_UF_ENCRYPTED_TEXT_PASSWORD_ALLOWED](https://msdn.microsoft.com/library/aa772300) 	The user can send an encrypted password.
	TEMP_DUPLICATE_ACCOUNT = 0x00000100  #[ADS_UF_TEMP_DUPLICATE_ACCOUNT](https://msdn.microsoft.com/library/aa772300) 	This is an account for users whose primary account is in another domain. This 	account provides user access to this domain, but not to any domain that trusts this domain. Also known as a local user account.	
	NORMAL_ACCOUNT = 0x00000200  #[ADS_UF_NORMAL_ACCOUNT](https://msdn.microsoft.com/library/aa772300) 	This is a default account type that represe	nts a typical user.	
	INTERDOMAIN_TRUST_ACCOUNT = 0x00000800  #[ADS_UF_INTERDOMAIN_TRUST_ACCOUNT](https://msdn.microsoft.com/library/aa772300) 	This is a permit to trust accou	nt for a system domain that trusts other do	mains.
	WORKSTATION_TRUST_ACCOUNT = 0x00001000  #[ADS_UF_WORKSTATION_TRUST_ACCOUNT](https://msdn.microsoft.com/library/aa772300) 	This is a computer account for 	a computer that is a member of this domain.	
	SERVER_TRUST_ACCOUNT = 0x00002000  #[ADS_UF_SERVER_TRUST_ACCOUNT](https://msdn.microsoft.com/library/aa772300) 	This is a computer account for a system	 backup domain controller that is a member 	of 	this domain.	
	NA_1 = 0x00004000 	#	N/A 	Not used.
	NA_2 = 0x00008000 	#	N/A 	Not used.
	DONT_EXPIRE_PASSWD = 0x00010000 	 #[ADS_UF_DONT_EXPIRE_PASSWD](https://msdn.microsoft.com/library/aa772300) 	The password for this account will never expire.
	MNS_LOGON_ACCOUNT = 0x00020000 	 #[ADS_UF_MNS_LOGON_ACCOUNT](https://msdn.microsoft.com/library/aa772300) 	This is an MNS logon account.
	SMARTCARD_REQUIRED = 0x00040000 	 #[ADS_UF_SMARTCARD_REQUIRED](https://msdn.microsoft.com/library/aa772300) 	The user must log on using a smart card.
	TRUSTED_FOR_DELEGATION = 0x00080000 	 #[ADS_UF_TRUSTED_FOR_DELEGATION](https://msdn.microsoft.com/library/aa772300) 	The service account (user or computer account), under which a service runs, is 	trusted for 	Kerberos delegation. Any such service can impersonate a client requesting the service.	
	NOT_DELEGATED = 0x00100000 	 #[ADS_UF_NOT_DELEGATED](https://msdn.microsoft.com/library/aa772300) 	The security c	ontext of the user will not be delegated to a service even if the service	 	account is s	et as trusted for Kerberos delegation.	
	USE_DES_KEY_ONLY = 0x00200000 	 #[ADS_UF_USE_DES_KEY_ONLY](https://msdn	.microsoft.com/library/aa772300) 	Restrict this principal to use only Data Encryption Standard (DES) encryption types for 	keys.	
	DONT_REQUIRE_PREAUTH = 0x00400000  #[ADS_UF_DONT_REQUIRE_PREAUTH](https://msdn.microsoft.com/library/aa772300) 	This account does not require Kerberos pre-authentication for logon.
	PASSWORD_EXPIRED = 0x00800000  #[ADS_UF_PASSWORD_EXPIRED](https://msdn.microsoft.com/library/aa772300) 	The user password has expired. This flag is created by the system using data from the [	Pwd-L	ast-Set](a-pwdlastset.md) attribute and the domain policy.	
	TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION = 0x01000000  #[ADS_UF_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION](https://msdn.microsoft.com/library/aa772300) 	The account is enabled for delegation. This is a security-sensi	tive 	setting; accounts with this option enabled should be strictly controlled. This setting enables a service running under the account to assume a client identity and authenticate 	as that user to other remote servers on the network.


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from msldap.commons.utils import bh_dt_convert
from msldap.commons.utils import FUNCTIONAL_LEVELS

MSADInfo_ATTRS = [
	'auditingPolicy', 'creationTime', 'dc', 'distinguishedName', 
	'forceLogoff', 'instanceType', 'lockoutDuration', 'lockOutObservationWindow', 
	'lockoutThreshold', 'masteredBy', 'maxPwdAge', 'minPwdAge', 'minPwdLength', 
	'name', 'nextRid', 'nTSecurityDescriptor', 'objectCategory', 'objectClass', 
	'objectGUID', 'objectSid', 'pwdHistoryLength', 
	'pwdProperties', 'serverState', 'systemFlags', 'uASCompat', 'uSNChanged', 
	'uSNCreated', 'whenChanged', 'whenCreated', 'rIDManagerReference',
	'msDS-Behavior-Version', 'description', 'isDeleted', 'gPLink', 'ms-DS-MachineAccountQuota'
]

class MSADInfo:
	def __init__(self):
		self.auditingPolicy = None #dunno
		self.creationTime = None #datetime
		self.dc = None #str
		self.distinguishedName = None #string
		self.forceLogoff = None #int
		self.instanceType = None #int
		self.lockoutDuration = None #int
		self.lockOutObservationWindow = None #int
		self.lockoutThreshold = None #int
		self.masteredBy = None #str
		self.maxPwdAge = None #int
		self.minPwdAge = None #int
		self.minPwdLength = None #int
		self.name = None #str
		self.nextRid = None #int
		self.nTSecurityDescriptor = None #str
		self.objectCategory = None #str
		self.objectClass = None #str
		self.objectGUID = None #str
		self.objectSid = None #str
		self.pwdHistoryLength = None #int
		self.pwdProperties = None #int
		self.serverState = None #int
		self.systemFlags = None #int
		self.uASCompat = None #int
		self.uSNChanged = None #int
		self.uSNCreated = None #int
		self.whenChanged = None #datetime
		self.whenCreated = None #datetime
		self.rIDManagerReference = None #str
		self.domainmodelevel = None
		self.description = None
		self.isDeleted = None
		self.gPLink = None
		self.machineAccountQuota = None
	
	@staticmethod
	def from_ldap(entry):
		adi = MSADInfo()
		adi.auditingPolicy = entry['attributes'].get('auditingPolicy') #dunno
		adi.creationTime = entry['attributes'].get('creationTime') #datetime
		adi.dc = entry['attributes'].get('dc') #str
		adi.distinguishedName = entry['attributes'].get('distinguishedName') #string
		adi.forceLogoff = entry['attributes'].get('forceLogoff') #int
		adi.instanceType = entry['attributes'].get('instanceType') #int
		adi.lockoutDuration = entry['attributes'].get('lockoutDuration') #int
		adi.lockOutObservationWindow = entry['attributes'].get('lockOutObservationWindow') #int
		adi.lockoutThreshold = entry['attributes'].get('lockoutThreshold') #int
		adi.masteredBy = entry['attributes'].get('masteredBy') #str
		adi.maxPwdAge = entry['attributes'].get('maxPwdAge') #int
		adi.minPwdAge = entry['attributes'].get('minPwdAge') #int
		adi.minPwdLength = entry['attributes'].get('minPwdLength') #int
		adi.name = entry['attributes'].get('name') #str
		adi.nextRid = entry['attributes'].get('nextRid') #int
		adi.nTSecurityDescriptor = entry['attributes'].get('nTSecurityDescriptor') #str
		adi.objectCategory = entry['attributes'].get('objectCategory') #str
		adi.objectClass = entry['attributes'].get('objectClass') #str
		adi.objectGUID = entry['attributes'].get('objectGUID') #str
		adi.objectSid = entry['attributes'].get('objectSid') #str
		adi.pwdHistoryLength = entry['attributes'].get('pwdHistoryLength') #int
		adi.pwdProperties = entry['attributes'].get('pwdProperties') #int
		adi.serverState = entry['attributes'].get('serverState') #int
		adi.systemFlags = entry['attributes'].get('systemFlags') #int
		adi.uASCompat = entry['attributes'].get('uASCompat') #int
		adi.uSNChanged = entry['attributes'].get('uSNChanged') #int
		adi.uSNCreated = entry['attributes'].get('uSNCreated') #int
		adi.whenChanged = entry['attributes'].get('whenChanged') #datetime
		adi.whenCreated = entry['attributes'].get('whenCreated') #datetime
		adi.rIDManagerReference = entry['attributes'].get('rIDManagerReference')
		adi.description = entry['attributes'].get('description')
		adi.isDeleted = entry['attributes'].get('isDeleted')
		adi.gPLink = entry['attributes'].get('gPLink')
		adi.machineAccountQuota = entry['attributes'].get('ms-DS-MachineAccountQuota')

		#https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-adts/564dc969-6db3-49b3-891a-f2f8d0a68a7f
		adi.domainmodelevel = entry['attributes'].get('msDS-Behavior-Version')

		return adi
	
	def to_dict(self):
		d = {}
		d['auditingPolicy'] = self.auditingPolicy
		d['creationTime'] = self.creationTime
		d['dc'] = self.dc
		d['distinguishedName'] = self.distinguishedName
		d['forceLogoff'] = self.forceLogoff
		d['instanceType'] = self.instanceType
		d['lockoutDuration'] = self.lockoutDuration
		d['lockOutObservationWindow'] = self.lockOutObservationWindow
		d['lockoutThreshold'] = self.lockoutThreshold
		d['masteredBy'] = self.masteredBy
		d['maxPwdAge'] = self.maxPwdAge
		d['minPwdAge'] = self.minPwdAge
		d['minPwdLength'] = self.minPwdLength
		d['name'] = self.name
		d['nextRid'] = self.nextRid
		d['nTSecurityDescriptor'] = self.nTSecurityDescriptor
		d['objectCategory'] = self.objectCategory
		d['objectClass'] = self.objectClass
		d['objectGUID'] = self.objectGUID
		d['objectSid'] = self.objectSid
		d['pwdHistoryLength'] = self.pwdHistoryLength
		d['pwdProperties'] = self.pwdProperties
		d['serverState'] = self.serverState
		d['systemFlags'] = self.systemFlags
		d['uASCompat'] = self.uASCompat
		d['uSNChanged'] = self.uSNChanged
		d['uSNCreated'] = self.uSNCreated
		d['whenChanged'] = self.whenChanged
		d['whenCreated'] = self.whenCreated
		d['domainmodelevel'] = self.domainmodelevel
		d['description'] = self.description
		d['ms-DS-MachineAccountQuota'] = self.machineAccountQuota
		return d


	def __str__(self):
		t = 'MSADInfo\n'
		t += 'auditingPolicy: %s\n' % self.auditingPolicy 
		t += 'creationTime: %s\n' % self.creationTime 
		t += 'dc: %s\n' % self.dc 
		t += 'distinguishedName: %s\n' % self.distinguishedName 
		t += 'forceLogoff: %s\n' % self.forceLogoff 
		t += 'instanceType: %s\n' % self.instanceType 
		t += 'lockoutDuration: %s\n' % self.lockoutDuration 
		t += 'lockOutObservationWindow: %s\n' % self.lockOutObservationWindow 
		t += 'lockoutThreshold: %s\n' % self.lockoutThreshold 
		t += 'masteredBy: %s\n' % self.masteredBy 
		t += 'maxPwdAge: %s\n' % self.maxPwdAge 
		t += 'minPwdAge: %s\n' % self.minPwdAge 
		t += 'minPwdLength: %s\n' % self.minPwdLength 
		t += 'name: %s\n' % self.name 
		t += 'nextRid: %s\n' % self.nextRid 
		t += 'nTSecurityDescriptor: %s\n' % self.nTSecurityDescriptor 
		t += 'objectCategory: %s\n' % self.objectCategory 
		t += 'objectClass: %s\n' % self.objectClass 
		t += 'objectGUID: %s\n' % self.objectGUID 
		t += 'objectSid: %s\n' % self.objectSid 
		t += 'pwdHistoryLength: %s\n' % self.pwdHistoryLength
		t += 'pwdProperties: %s\n' % self.pwdProperties
		t += 'serverState: %s\n' % self.serverState
		t += 'systemFlags: %s\n' % self.systemFlags
		t += 'uASCompat: %s\n' % self.uASCompat
		t += 'uSNChanged: %s\n' % self.uSNChanged
		t += 'uSNCreated: %s\n' % self.uSNCreated
		t += 'whenChanged: %s\n' % self.whenChanged
		t += 'whenCreated: %s\n' % self.whenCreated
		t += 'domainmodelevel: %s\n' % self.domainmodelevel
		t += 'description: %s\n' % self.description
		t += 'isDeleted: %s\n' % self.isDeleted
		t += 'gPLink: %s\n' % self.gPLink
		t += 'ms-DS-MachineAccountQuota: %s\n' % self.machineAccountQuota
		return t
	
	def get_row(self, attrs):
		t = self.to_dict()
		return [str(t.get(x)) for x in attrs]
	
	def to_bh(self, domain):
		try:
			functional_level = FUNCTIONAL_LEVELS[int(self.domainmodelevel)]
		except KeyError:
			functional_level = 'Unknown'
		return {
			'_gPLink' : self.gPLink, #always empty
			'Aces' : [],
			'ObjectIdentifier' : str(self.objectSid),
			"IsDeleted": bool(self.isDeleted),
			"IsACLProtected": False , # Post processing
			"Trusts" : [], #Post processing
			"ChildObjects": [], #Post processing
			"GPOChanges" : {
				"LocalAdmins": [],
				"RemoteDesktopUsers": [],
				"DcomUsers": [],
				"PSRemoteUsers": [],
				"AffectedComputers": []
			},
			"Links": [], # Post processing
			'Properties' : {
				'name' : domain,
				'domain' : domain,
				'domainsid' : str(self.objectSid).rsplit('-',1)[0] , 
				'distinguishedname' : str(self.distinguishedName).upper(), 
				'description' : self.description,
				'functionallevel' : functional_level,
				'highvalue' : True,
				'whencreated' : bh_dt_convert(self.whenCreated),
			}
			
		}


# parts of the dns implementation was taken from https://github.com/dirkjanm/adidnsdump

import datetime
import enum
import io
from typing import List

from msldap.commons.utils import bytes2ipv4, bytes2ipv6


class DNS_RECORD_TYPE(enum.Enum):
	ZERO = 0x0000 #An empty record type ([RFC1034] section 3.6 and [RFC1035] section 3.2.2).
	A = 0x0001 #An A record type, used for storing an IP address ([RFC1035] section 3.2.2).
	NS = 0x0002 #An authoritative name-server record type ([RFC1034] section 3.6 and [RFC1035] section 3.2.2).
	MD = 0x0003 #A mail-destination record type ([RFC1035] section 3.2.2).
	MF = 0x0004 #A mail forwarder record type ([RFC1035] section 3.2.2).
	CNAME = 0x0005 #A record type that contains the canonical name of a DNS alias ([RFC1035] section 3.2.2).
	SOA = 0x0006 #A Start of Authority (SOA) record type ([RFC1035] section 3.2.2).
	MB = 0x0007 #A mailbox record type ([RFC1035] section 3.2.2).
	MG = 0x0008 #A mail group member record type ([RFC1035] section 3.2.2).
	MR = 0x0009 #A mail-rename record type ([RFC1035] section 3.2.2).
	NULL = 0x000A #A record type for completion queries ([RFC1035] section 3.2.2).
	WKS = 0x000B #A record type for a well-known service ([RFC1035] section 3.2.2).
	PTR = 0x000C #A record type containing FQDN pointer ([RFC1035] section 3.2.2).
	HINFO = 0x000D #A host information record type ([RFC1035] section 3.2.2).
	MINFO = 0x000E #A mailbox or mailing list information record type ([RFC1035] section 3.2.2).
	MX = 0x000F #A mail-exchanger record type ([RFC1035] section 3.2.2).
	TXT = 0x0010 #A record type containing a text string ([RFC1035] section 3.2.2).
	RP = 0x0011 #A responsible-person record type [RFC1183].
	AFSDB = 0x0012 #A record type containing AFS database location [RFC1183].
	X25 = 0x0013 #An X25 PSDN address record type [RFC1183].
	ISDN = 0x0014 #An ISDN address record type [RFC1183].
	RT = 0x0015 #A route through record type [RFC1183].
	SIG = 0x0018 #A cryptographic public key signature record type [RFC2931].
	KEY = 0x0019 #A record type containing public key used in DNSSEC [RFC2535].
	AAAA = 0x001C #An IPv6 address record type [RFC3596].
	LOC = 0x001D #A location information record type [RFC1876].
	NXT = 0x001E #A next-domain record type [RFC2065].
	SRV = 0x0021 #A server selection record type [RFC2782].
	ATMA = 0x0022 # An Asynchronous Transfer Mode (ATM) address record type [ATMA].
	NAPTR = 0x0023 #An NAPTR record ype [RFC2915].
	DNAME = 0x0027 #A DNAME record type [RFC2672].
	DS = 0x002B #A DS record type [RFC4034].
	RRSIG = 0x002E #An RRSIG record type [RFC4034].
	NSEC = 0x002F #An NSEC record type [RFC4034].
	DNSKEY = 0x0030 #A DNSKEY record type [RFC4034].
	DHCID = 0x0031 #A DHCID record type [RFC4701]. 
	NSEC3 = 0x0032 #An NSEC3 record type [RFC5155].
	NSEC3PARAM = 0x0033 #An NSEC3PARAM record type [RFC5155].
	TLSA = 0x0034 #A TLSA record type [RFC6698].
	ALL = 0x00FF #A query-only type requesting all records [RFC1035].
	WINS = 0xFF01 # A record type containing Windows Internet Name Service (WINS) forward lookup data [MS-WINSRA].
	WINSR = 0xFF02 #A record type containing WINS reverse lookup data [MS-WINSRA].


# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-dnsp/ac793981-1c60-43b8-be59-cdbb5c4ecb8a
class DNS_RECORD:
	def __init__(self):
		self.DataLength:int = None
		self.Type:DNS_RECORD_TYPE = None
		self.Version: bytes = b'\x05'
		self.Rank:bytes = None
		self.Flags:bytes = b'\x00\x00'
		self.Serial:bytes = None
		self.TtlSeconds:int = None
		self.Reserved:bytes = b'\x00\x00\x00\x00'
		self.TimeStamp:bytes = None
		self.Data:bytes = None
	
	@staticmethod
	def from_bytes(data):
		return DNS_RECORD.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		res = DNS_RECORD()
		res.DataLength = int.from_bytes(buff.read(2), 'little', signed = False)
		res.Type = DNS_RECORD_TYPE(int.from_bytes(buff.read(2), 'little', signed = False))
		res.Version = buff.read(1)
		res.Rank = buff.read(1)
		res.Flags = buff.read(2)
		res.Serial = buff.read(4)
		res.TtlSeconds = int.from_bytes(buff.read(4), 'little', signed = False)
		res.Reserved = buff.read(4)
		res.TimeStamp = buff.read(4)
		res.Data = buff.read(res.DataLength)
		return res
	
	def get_formatted(self):
		if self.Data is None or len(self.Data) == 0:
			return None
		if self.Type in MSLDAP_DNS_TYPE_TO_CLASS:
			return MSLDAP_DNS_TYPE_TO_CLASS[self.Type].from_bytes(self.Data)
		return DNS_RPC_RECORD_UNKNOWN.from_bytes(self.Data)
	
	def __str__(self):
		t = '==== DNS_RECORD ====\r\n'
		t += 'DataLength: %s\r\n' % self.DataLength
		t += 'Type: %s\r\n' % self.Type.name
		t += 'Version: %s\r\n' % self.Version.hex()
		t += 'Rank: %s\r\n' % self.Rank.hex()
		t += 'Flags: %s\r\n' % self.Flags.hex()
		t += 'Serial: %s\r\n' % self.Serial.hex()
		t += 'TtlSeconds: %s\r\n' % self.TtlSeconds
		t += 'Reserved: %s\r\n' % self.Reserved.hex()
		t += 'TimeStamp: %s\r\n' % self.TimeStamp.hex()
		t += 'Data: %s\r\n' % self.Data.hex()
		return t
	
class DNS_RPC_RECORD_UNKNOWN:
	def __init__(self):
		self.Data:bytes = None
	
	@staticmethod
	def from_bytes(data):
		return DNS_RPC_RECORD_UNKNOWN.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		res = DNS_RPC_RECORD_UNKNOWN()
		res.Data = buff.read()
		return res

	def to_line(self, separator = '\t'):
		return str(self)
	
	def __str__(self):
		return self.Data.hex()

class DNS_RPC_RECORD_A:
	def __init__(self):
		self.IpAddress:str = None
	
	@staticmethod
	def from_bytes(data):
		return DNS_RPC_RECORD_A.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		res = DNS_RPC_RECORD_A()
		#webassembly...
		#res.IpAddress = socket.inet_ntoa(buff.read(4))
		ipv4_bytes = buff.read(4)
		
		# Convert each byte to an integer and join them with dots to form the IPv4 address
		res.IpAddress = bytes2ipv4(ipv4_bytes)
		
		return res
	
	def to_line(self, separator = '\t'):
		return str(self)
	
	def __str__(self):
		return self.IpAddress

class DNS_RPC_RECORD_AAAA:
	def __init__(self):
		self.IpAddress:str = None
	
	@staticmethod
	def from_bytes(data):
		return DNS_RPC_RECORD_AAAA.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		res = DNS_RPC_RECORD_AAAA()
		# webassembly pyodide has problems with socket implementation
		#res.IpAddress = socket.inet_ntop(socket.AF_INET6, buff.read(16))
		ipv6_bytes = buff.read(16)

		# Convert the hex string into the standard IPv6 format
		res.IpAddress = bytes2ipv6(ipv6_bytes)
		return res
	
	def to_line(self, separator = '\t'):
		return str(self)
	
	def __str__(self):
		return self.IpAddress

class DNS_RPC_RECORD_SRV:
	def __init__(self):
		self.wPriority:int = None
		self.wWeight:int = None
		self.wPort:int = None
		self.nameTarget:DNS_COUNT_NAME = None
	
	@staticmethod
	def from_bytes(data):
		return DNS_RPC_RECORD_SRV.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		res = DNS_RPC_RECORD_SRV()
		res.wPriority = int.from_bytes(buff.read(2), 'little', signed = False)
		res.wWeight = int.from_bytes(buff.read(2), 'little', signed = False)
		res.wPort = int.from_bytes(buff.read(2), 'little', signed = False)
		res.nameTarget = DNS_COUNT_NAME.from_buffer(buff)
		return res
	
	def to_line(self, separator = '\t'):
		return separator.join([
			str(self.wPriority), 
			str(self.wWeight), 
			str(self.wPort), 
			str(self.nameTarget)
		])
	
	def __str__(self):
		return 'Prio: %s | Weight: %s | Port: %s | Name: %s' % (self.wPriority, self.wWeight, self.wPort, self.nameTarget)

class DNS_RPC_RECORD_TS:
	def __init__(self):
		self.entombedTime:datetime.datetime = None
	
	@staticmethod
	def from_bytes(data):
		return DNS_RPC_RECORD_TS.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		res = DNS_RPC_RECORD_TS()
		tsraw = buff.read(8, signed = False, byteorder = 'little')
		microseconds = tsraw / 10
		try:
			res.entombedTime = datetime.datetime(1601,1,1) + datetime.timedelta(microseconds=microseconds)
		except OverflowError:
			return None
		return res
	
	def to_line(self, separator = '\t'):
		return str(self)
	
	def __str__(self):
		return str(self.entombedTime)

class DNS_COUNT_NAME:
	def __init__(self):
		self.Length:int = None
		self.LabelCount:int = None
		#self.RawName:str = None
		self.labels:List[str] = []
	
	@staticmethod
	def from_bytes(data):
		return DNS_COUNT_NAME.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		res = DNS_COUNT_NAME()
		res.Length = int.from_bytes(buff.read(1), 'little', signed = False)
		res.LabelCount = int.from_bytes(buff.read(1), 'little', signed = False)
		#res.RawName = buff.read(res.Length)
		for _ in range(res.LabelCount):
			res.labels.append(buff.read(buff.read(1)[0]).decode('utf-8'))
		buff.read(1) # IMPORTANT! The last byte is always 0x00 and is not indicated in the LabelCount
		return res
	
	def to_line(self, separator = '\t'):
		return str(self)
	
	def __str__(self):
		return '.'.join(self.labels)

class DNS_RPC_RECORD_NULL:
	def __init__(self):
		self.bData:bytes = b''
	
	@staticmethod
	def from_bytes(data):
		return DNS_RPC_RECORD_NULL.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		res = DNS_RPC_RECORD_NULL()
		return res
	
	def to_line(self, separator = '\t'):
		return ''
	
	def __str__(self):
		return ''


class DNS_RPC_RECORD_SOA:
	def __init__(self):
		self.dwSerialNo:int = None
		self.dwRefresh:int = None
		self.dwRetry:int = None
		self.dwExpire:int = None
		self.dwMinimumTtl:int = None
		self.namePrimaryServer:DNS_COUNT_NAME = None
		self.zoneAdminEmail:DNS_COUNT_NAME = None
	
	@staticmethod
	def from_bytes(data):
		return DNS_RPC_RECORD_SOA.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		res = DNS_RPC_RECORD_SOA()
		res.dwSerialNo = int.from_bytes(buff.read(4), 'little', signed = False)
		res.dwRefresh = int.from_bytes(buff.read(4), 'little', signed = False)
		res.dwRetry = int.from_bytes(buff.read(4), 'little', signed = False)
		res.dwExpire = int.from_bytes(buff.read(4), 'little', signed = False)
		res.dwMinimumTtl = int.from_bytes(buff.read(4), 'little', signed = False)
		res.namePrimaryServer = DNS_COUNT_NAME.from_buffer(buff)
		res.zoneAdminEmail = DNS_COUNT_NAME.from_buffer(buff)
		return res
	
	def to_line(self, separator = '\t'):
		return separator.join([
			str(self.dwSerialNo), 
			str(self.dwRefresh), 
			str(self.dwRetry), 
			str(self.dwExpire), 
			str(self.dwMinimumTtl),
			str(self.namePrimaryServer), 
			str(self.zoneAdminEmail)
		])
	
	def __str__(self):
		return 'Serial: %s | Refresh: %s | Retry: %s | Expire: %s | TTL: %s | NS: %s | Email: %s' % (self.dwSerialNo, self.dwRefresh, self.dwRetry, self.dwExpire, self.dwMinimumTtl, self.namePrimaryServer, self.zoneAdminEmail)



MSLDAP_DNS_TYPE_TO_CLASS = {
	DNS_RECORD_TYPE.ZERO : DNS_RPC_RECORD_NULL,
	DNS_RECORD_TYPE.A : DNS_RPC_RECORD_A,
	DNS_RECORD_TYPE.AAAA : DNS_RPC_RECORD_AAAA,
	DNS_RECORD_TYPE.SRV : DNS_RPC_RECORD_SRV,
	DNS_RECORD_TYPE.PTR : DNS_COUNT_NAME,
	DNS_RECORD_TYPE.NS : DNS_COUNT_NAME,
	DNS_RECORD_TYPE.CNAME : DNS_COUNT_NAME,
	DNS_RECORD_TYPE.DNAME : DNS_COUNT_NAME,
	DNS_RECORD_TYPE.MB : DNS_COUNT_NAME,
	DNS_RECORD_TYPE.MR : DNS_COUNT_NAME,
	DNS_RECORD_TYPE.MG : DNS_COUNT_NAME,
	DNS_RECORD_TYPE.MD : DNS_COUNT_NAME,
	DNS_RECORD_TYPE.MF : DNS_COUNT_NAME,
	DNS_RECORD_TYPE.SOA : DNS_RPC_RECORD_SOA,
}



from asysocks.unicomm.common.target import UniTarget, UniProto
from asysocks.unicomm.server import UniServer
from msldap.network.packetizer import LDAPPacketizer
from msldap.relay.serverconnection import LDAPRelayServerConnection
from asyauth.protocols.spnego.relay.native import spnegorelay_ntlm_factory
from asyauth.protocols.ntlm.relay.native import NTLMRelaySettings, ntlmrelay_factory
import traceback
import asyncio

class LDAPServerSettings:
	def __init__(self, gssapi_factory):
		self.gssapi_factory = gssapi_factory
	
	@property
	def gssapi(self):
		return self.gssapi_factory()
	
class LDAPRelayServer:
	def __init__(self, target, settings):
		self.target = target
		self.settings = settings
		self.server = None
		self.serving_task = None
		self.connections = {}
		self.conn_ctr = 0
		
	def get_ctr(self):
		self.conn_ctr += 1
		return self.conn_ctr

	async def __handle_connection(self):
		try:
			async for connection in self.server.serve():
				print('connection in!')
				
				smbconnection = LDAPRelayServerConnection(self.settings, connection)
				self.connections[self.get_ctr()] = smbconnection
				x = asyncio.create_task(smbconnection.run())

		except Exception as e:
			traceback.print_exc()
			return

	async def run(self):
		self.server = UniServer(self.target, LDAPPacketizer())
		self.serving_task = asyncio.create_task(self.__handle_connection())
		return self.serving_task

async def test_relay_queue(rq):
	try:
		from aiosmb.connection import SMBConnection
		from aiosmb.commons.connection.target import SMBTarget
		from aiosmb.commons.interfaces.machine import SMBMachine
		test_target = SMBTarget('10.10.10.2')
		while True:
			item = await rq.get()
			print(item)
			connection = SMBConnection(item, test_target, preserve_gssapi=False, nosign=True)
			_, err = await connection.login()
			if err is not None:
				print('SMB client login err: %s' % err)
				print(traceback.format_tb(err.__traceback__))
				continue
			machine = SMBMachine(connection)
			async for share, err in machine.list_shares():
				if err is not None:
					print('SMB client list_shares err: %s' % err)
					continue
				print(share)

	except Exception as e:
		traceback.print_exc()
		return

async def amain():
	try:
		auth_relay_queue = asyncio.Queue()
		x = asyncio.create_task(test_relay_queue(auth_relay_queue))
		target = UniTarget('0.0.0.0', 636, UniProto.SERVER_SSL_TCP)

		settings = LDAPServerSettings(lambda: spnegorelay_ntlm_factory(auth_relay_queue, lambda: ntlmrelay_factory()))
		server = LDAPRelayServer(target, settings)
		server_task = await server.run()
		await server_task
	except Exception as e:
		traceback.print_exc()
		return
if __name__ == '__main__':
	asyncio.run(amain())


from msldap.protocol.utils import calcualte_length
from msldap.protocol.messages import LDAPMessage, AuthenticationChoice, protocolOp, BindResponse
from msldap import logger
import traceback
import asyncio

class LDAPRelayServerConnection:
	def __init__(self, settings, connection):
		self.settings = settings
		self.gssapi = settings.gssapi
		self.ntlm = self.gssapi.authentication_contexts['NTLMSSP - Microsoft NTLM Security Support Provider']
		self.connection = connection
		self.__auth_type = None
		self.__auth_iter = 0
	
	async def log_async(self, level, msg):
		if self.settings.log_q is not None:
			src = 'LDAPCON-%s:%s' % (self.client_ip, self.client_port)
			await self.settings.log_q.put((src, level, msg))
		else:
			logger.log(level, msg)

	async def terminate(self):
		self.handle_in_task.cancel()
	
	async def __handle_ldap_in(self):
		async for msg_data in self.connection.read():
			try:
				msg = LDAPMessage.load(msg_data)
				if msg['protocolOp']._choice == 0:
					await self.__bindreq(msg)
				else:
					raise Exception('Unknown LDAP message! %s' % msg.native)

			except Exception as e:
				await self.log_async(1, str(e))
				return

	async def __bindreq(self, msg):
		try:
			msg_id = msg.native['messageID']
			authdata_raw = msg.native['protocolOp']['authentication']
			if isinstance(authdata_raw, bytes) is True:
				self.__auth_type = 'NTLM'
				if self.__auth_iter == 0:
					t = {
						'resultCode' : 0,
						'matchedDN' : 'NTLM'.encode(),
						'diagnosticMessage' : b'',		
					}
					po = {'bindResponse' : BindResponse(t)}
					b= {
						'messageID' : msg_id,
						'protocolOp' : protocolOp(po),					
					}
					resp = LDAPMessage(b)
					await self.connection.write(resp.dump())
					self.__auth_iter += 1
					return
					
				else:
					resdata, to_conitnue, err = await self.ntlm.authenticate_relay_server(authdata_raw)
					if err is not None:
						raise err
					
					if resdata is None:
						t = {
						'resultCode' : 49,
						'matchedDN' : b'',
						'diagnosticMessage' : b'8009030C: LdapErr: DSID-0C090569, comment: AcceptSecurityContext error, data 52e, v4563\x00',		
						}
						po = {'bindResponse' : BindResponse(t)}
						b= {
							'messageID' : msg_id,
							'protocolOp' : protocolOp(po),
						}
						resp = LDAPMessage(b)

						await self.connection.write(resp.dump())

						await self.terminate()
						return
					
					t = {
						'resultCode' : 0,
						'matchedDN' : resdata,
						'diagnosticMessage' : b'',		
					}
					po = {'bindResponse' : BindResponse(t)}
					b= {
						'messageID' : msg_id,
						'protocolOp' : protocolOp(po),
					}
					resp = LDAPMessage(b)
					await self.connection.write(resp.dump())
					self.__auth_iter += 1
					return

			if isinstance(authdata_raw, dict) is True:
				self.__auth_type = 'GSSAPI'
				resdata, to_conitnue, err = await self.gssapi.authenticate_relay_server(authdata_raw['credentials'])
				if err is not None:
					raise err
				
				if resdata is None:
					t = {
						'resultCode' : 49,
						'matchedDN' : b'',
						'diagnosticMessage' : b'8009030C: LdapErr: DSID-0C090569, comment: AcceptSecurityContext error, data 52e, v4563\x00',		
					}
					po = {'bindResponse' : BindResponse(t)}
					b= {
						'messageID' : msg_id,
						'protocolOp' : protocolOp(po),
					}
					resp = LDAPMessage(b)

					await self.connection.write(resp.dump())

					await self.terminate()
					return

				t = {
					'resultCode' : 14,
					'matchedDN' : b'',
					'diagnosticMessage' : b'',
					'serverSaslCreds': 	resdata
				}
				po = {'bindResponse' : BindResponse(t)}
				b= {
					'messageID' : msg_id,
					'protocolOp' : protocolOp(po),					
				}
				resp = LDAPMessage(b)
				await self.connection.write(resp.dump())
				self.__auth_iter += 1
			
			else:
				raise Exception('Unknown auth method %s' % authdata_raw)

				
		except Exception as e:
			traceback.print_exc()
			return
	
	async def run(self):
		self.handle_in_task = asyncio.create_task(self.__handle_ldap_in())
		await self.handle_in_task


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import asyncio
import traceback
import logging

from msldap import logger
from asysocks import logger as sockslogger
from msldap.commons.factory import LDAPConnectionFactory

class MSLDAPCompDomainList:
	def __init__(self, ldap_url):
		self.conn_url = ldap_url
		self.connection = None
		self.adinfo = None
		self.ldapinfo = None
		self.domain_name = None

	async def login(self):
		"""Performs connection and login"""
		try:
			logger.debug(self.conn_url.get_credential())
			logger.debug(self.conn_url.get_target())
			
			
			self.connection = self.conn_url.get_client()
			_, err = await self.connection.connect()
			if err is not None:
				raise err
			
			return True, None
		except Exception as e:
			return False, e

	async def do_adinfo(self, show = True):
		"""Prints detailed Active Driectory info"""
		try:
			if self.adinfo is None:
				self.adinfo = self.connection._ldapinfo
				self.domain_name = self.adinfo.distinguishedName.replace('DC','').replace('=','').replace(',','.')
			if show is True:
				print(self.adinfo)
			
			return True, None
		except Exception as e:
			return False, e

	async def run(self):
		try:
			_, err = await self.login()
			if err is not None:
				raise err
			_, err = await self.do_adinfo(False)
			if err is not None:
				raise err
		
			async for machine, err in self.connection.get_all_machines(attrs=['sAMAccountName', 'dNSHostName']):
				if err is not None:
					raise err
					
				dns = machine.dNSHostName
				if dns is None:
					dns = '%s.%s' % (machine.sAMAccountName[:-1], self.domain_name)

				print(str(dns))
		except:
			traceback.print_exc()


def main():
	import argparse
	parser = argparse.ArgumentParser(description='MS LDAP library')
	parser.add_argument('-v', '--verbose', action='count', default=0, help='Verbosity, can be stacked')
	parser.add_argument('-n', '--no-interactive', action='store_true')
	parser.add_argument('url', help='Connection string in URL format.')

	args = parser.parse_args()
	
	
	###### VERBOSITY
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	else:
		sockslogger.setLevel(logging.DEBUG)
		logger.setLevel(logging.DEBUG)
		logging.basicConfig(level=logging.DEBUG)

	ldap_url = LDAPConnectionFactory.from_url(args.url)
	compdomlist = MSLDAPCompDomainList(ldap_url)


	asyncio.run(compdomlist.run())

if __name__ == '__main__':
	main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .drawing import BoxStyle
from .traversal import DictTraversal
from .util import KeyArgsConstructor


class LeftAligned(KeyArgsConstructor):
    """Creates a renderer for a left-aligned tree.

    Any attributes of the resulting class instances can be set using
    constructor arguments."""

    draw = BoxStyle()
    "The draw style used. See :class:`~asciitree.drawing.Style`."
    traverse = DictTraversal()
    "Traversal method. See :class:`~asciitree.traversal.Traversal`."

    def render(self, node):
        """Renders a node. This function is used internally, as it returns
        a list of lines. Use :func:`~asciitree.LeftAligned.__call__` instead.
        """
        lines = []

        children = self.traverse.get_children(node)
        lines.append(self.draw.node_label(self.traverse.get_text(node)))

        for n, child in enumerate(children):
            child_tree = self.render(child)

            if n == len(children) - 1:
                # last child does not get the line drawn
                lines.append(self.draw.last_child_head(child_tree.pop(0)))
                lines.extend(self.draw.last_child_tail(l)
                             for l in child_tree)
            else:
                lines.append(self.draw.child_head(child_tree.pop(0)))
                lines.extend(self.draw.child_tail(l)
                             for l in child_tree)

        return lines

    def __call__(self, tree):
        """Render the tree into string suitable for console output.

        :param tree: A tree."""
        return '\n'.join(self.render(self.traverse.get_root(tree)))


# legacy support below

from .drawing import Style
from .traversal import Traversal


class LegacyStyle(Style):
    def node_label(self, text):
        return text

    def child_head(self, label):
        return '  +--' + label

    def child_tail(self, line):
        return '  |' + line

    def last_child_head(self, label):
        return '  +--' + label

    def last_child_tail(self, line):
        return '   ' + line


def draw_tree(node,
              child_iter=lambda n: n.children,
              text_str=str):
    """Support asciitree 0.2 API.

    This function solely exist to not break old code (using asciitree 0.2).
    Its use is deprecated."""
    return LeftAligned(traverse=Traversal(get_text=text_str,
                                          get_children=child_iter),
                       draw=LegacyStyle())(node)



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from msldap.wintypes import *
from msldap.ldap_objects.common import MSLDAP_UAC, vn
from winacl.dtyp.sid import SID
from msldap.commons.utils import bh_dt_convert


MSADGroup_ATTRS = [ 	
	'cn', 'distinguishedName', 'objectGUID', 'objectSid', 'groupType', 
	'instanceType', 'name', 'member', 'sAMAccountName', 'systemFlags', 
	'whenChanged', 'whenCreated', 'description', 'nTSecurityDescriptor',
	'sAMAccountType', 'adminCount', 'isDeleted'
]

MSADGroup_highvalue = ["S-1-5-32-544", "S-1-5-32-550", "S-1-5-32-549", "S-1-5-32-551", "S-1-5-32-548"]



class MSADGroup:
	def __init__(self):
		self.cn = None #str
		self.distinguishedName = None #dn
		self.objectGUID = None
		self.objectSid = None
		self.description = None
		self.groupType = None
		self.instanceType = None
		self.name = None
		self.member = None
		self.nTSecurityDescriptor = None
		self.sAMAccountName = None
		self.sAMAccountType = None
		self.systemFlags = None
		self.whenChanged = None
		self.whenCreated = None
		self.adminCount = None
		self.isDeleted = None
		
	def to_dict(self):
		d = {}
		d['cn'] = self.cn
		d['distinguishedName'] = self.distinguishedName
		d['objectGUID'] = self.objectGUID
		d['objectSid'] = self.objectSid
		d['description'] = self.description
		d['groupType'] = self.groupType
		d['instanceType'] = self.instanceType
		d['name'] = self.name
		d['member'] = self.member
		d['nTSecurityDescriptor'] = self.nTSecurityDescriptor
		d['sAMAccountName'] = self.sAMAccountName
		d['sAMAccountType'] = self.sAMAccountType
		d['systemFlags'] = self.systemFlags
		d['whenChanged'] = self.whenChanged
		d['whenCreated'] = self.whenCreated
		d['adminCount'] = self.adminCount
		d['isDeleted'] = self.isDeleted

		return d	
	
	@staticmethod
	def from_ldap(entry):
		t = MSADGroup()
		t.cn = entry['attributes'].get('cn')
		t.distinguishedName = entry['attributes'].get('distinguishedName')
		t.objectGUID = entry['attributes'].get('objectGUID')
		t.objectSid = entry['attributes'].get('objectSid')
		t.groupType = entry['attributes'].get('groupType')
		t.instanceType = entry['attributes'].get('instanceType')
		t.name = entry['attributes'].get('name')
		t.member = entry['attributes'].get('member')
		t.sAMAccountName = entry['attributes'].get('sAMAccountName')
		t.systemFlags = entry['attributes'].get('systemFlags')
		t.whenChanged = entry['attributes'].get('whenChanged')
		t.whenCreated = entry['attributes'].get('whenCreated')
		t.adminCount = entry['attributes'].get('adminCount')
		t.isDeleted = entry['attributes'].get('isDeleted')
		
		t.description =  entry['attributes'].get('description')
		if isinstance(t.description, list):
			if len(t.description) == 1:
				t.description = t.description[0]
			else:
				t.description = ', '.join(t.description)
		
		
		#temp = entry['attributes'].get('nTSecurityDescriptor')
		#if temp:
		#	t.nTSecurityDescriptor = SID.from_bytes(temp)
		return t
	
	def get_row(self, attrs):
		t = self.to_dict()
		return [str(t.get(x)) if x[:4]!='UAC_' else str(self.uac_to_textflag(x)) for x in attrs]

	def __str__(self):
		t = 'MSADGroup\r\n'
		for x in self.__dict__:
			if not isinstance(self.__dict__[x], (list, dict)):
				t += '%s: %s\r\n' % (x, str(self.__dict__[x]))
			else:
				t += '%s: %s\r\n' % (x, self.__dict__[x])
		return t

	def to_bh(self, domain):
		# Thx Dirk-jan
		def is_highvalue(sid:str):
			if sid.endswith("-512") or sid.endswith("-516") or sid.endswith("-519") or sid.endswith("-520"):
				return True
			if sid in MSADGroup_highvalue:
				return True
			return False

		return {
			'Aces' : [],
			'Members': [],
			'ObjectIdentifier' : self.objectSid,
			"IsDeleted": bool(self.isDeleted),
			"IsACLProtected": False , # Post processing
			'Properties' : {
				'name' : '%s@%s' % (self.name.upper(), domain.upper()),
				'domain' : domain,
				'domainsid' : str(self.objectSid).rsplit('-',1)[0] , 
				'distinguishedname' : str(self.distinguishedName).upper(), 
				'highvalue' : is_highvalue(str(self.objectSid)),
				'admincount' : bool(self.adminCount),
				'description' : self.description ,
				'samaccountname' : self.sAMAccountName ,
				'whencreated' : bh_dt_convert(self.whenCreated),
			},
		}



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from os import stat
from sqlite3 import connect
from asysocks.unicomm.common.target import UniTarget, UniProto
from urllib.parse import urlparse, parse_qs, unquote
from asysocks.unicomm.utils.paramprocessor import str_one, int_one, bool_one

msldaptarget_url_params = {
	'pagesize' : int_one,
	'rate' : int_one,
}


class MSLDAPTarget(UniTarget):
	"""
	Describes the connection to the server.
	
	:param host: IP address or hostname of the server
	:type host: str
	:param port: port of the LDAP service running on the server
	:type port: int
	:param proto: Connection protocol to be used
	:type proto: :class:`UniProto`
	:param tree: The tree to connect to
	:type tree: str
	:param proxies: specifies what kind of proxy to be used
	:type proxies: :class:`List[UniProxyTarget]`
	:param timeout: connection timeout in seconds
	:type timeout: int
	:param ldap_query_page_size: Maximum number of elements to fetch in each paged_query call.
	:type ldap_query_page_size: int
	:param ldap_query_ratelimit: rate limit of paged queries. This will cause a sleep (in seconds) between fetching of each page of the query
	:type ldap_query_ratelimit: float
	:param dc_ip: Ip address of the kerberos server (if kerberos is used)
	:type dc_ip: str
	"""
	def __init__(self, ip, port = 389, protocol = UniProto.CLIENT_TCP, tree = None, proxies = None, timeout = 10, ldap_query_page_size = 1000, ldap_query_ratelimit = 0, dns:str=None, dc_ip:str = None, domain:str = None, hostname:str = None, ssl_ctx = None):
		UniTarget.__init__(self, ip, port, protocol, timeout, hostname = hostname, ssl_ctx= ssl_ctx, proxies = proxies, domain = domain, dc_ip = dc_ip, dns=dns)
		self.tree = tree
		self.ldap_query_page_size = ldap_query_page_size
		self.ldap_query_ratelimit = ldap_query_ratelimit
	
	def to_target_string(self):
		return 'ldap/%s@%s' % (self.get_hostname_or_ip(), self.domain)  #ldap/WIN2019AD.test.corp @ TEST.CORP

	def get_host(self):
		if self.protocol == UniProto.CLIENT_SSL_TCP:
			proto = 'ldaps'
		elif self.protocol == UniProto.CLIENT_TCP:
			proto = 'ldap'
		return '%s://%s:%s' % (proto, self.get_hostname_or_ip(), self.port)

	def is_ssl(self):
		return self.protocol == UniProto.CLIENT_SSL_TCP
	
	@staticmethod
	def from_url(connection_url):
		url_e = urlparse(connection_url)
		url_dict = url_e._asdict()
		for prop, val in url_dict.items():
			if type(val) is str:
				url_dict[prop] = unquote(val)
		url_e = url_e._replace(**url_dict)
		schemes = []
		for item in url_e.scheme.upper().split('+'):
			schemes.append(item.replace('-','_'))
		if schemes[0] == 'LDAP':
			protocol = UniProto.CLIENT_TCP
			port = 389
		elif schemes[0] == 'LDAPS':
			protocol = UniProto.CLIENT_SSL_TCP
			port = 636
		elif schemes[0] == 'LDAP_SSL':
			protocol = UniProto.CLIENT_SSL_TCP
			port = 636
		elif schemes[0] == 'LDAP_TCP':
			protocol = UniProto.CLIENT_TCP
			port= 389
		elif schemes[0] == 'LDAP_UDP':
			raise NotImplementedError()
			protocol = UniProto.CLIENT_UDP
			port = 389
		elif schemes[0] == 'GC':
			protocol = UniProto.CLIENT_TCP
			port = 3268
		elif schemes[0] == 'GC_SSL':
			protocol = UniProto.CLIENT_SSL_TCP
			port = 3269
		else:
			raise Exception('Unknown protocol! %s' % schemes[0])
		
		if url_e.port:
			port = url_e.port
		if port is None:
			raise Exception('Port must be provided!')
		
		path = None
		if url_e.path not in ['/', '', None]:
			path = url_e.path
		
		unitarget, extraparams = UniTarget.from_url(connection_url, protocol, port, msldaptarget_url_params)
		pagesize = extraparams['pagesize'] if extraparams['pagesize'] is not None else 1000
		rate = extraparams['rate'] if extraparams['rate'] is not None else 0

		target = MSLDAPTarget(
			unitarget.ip, 
			port = unitarget.port, 
			protocol = unitarget.protocol, 
			tree = path, 
			proxies = unitarget.proxies, 
			timeout = unitarget.timeout, 
			ldap_query_page_size = pagesize, 
			ldap_query_ratelimit = rate,
			dns = unitarget.dns, 
			dc_ip = unitarget.dc_ip, 
			domain = unitarget.domain, 
			hostname = unitarget.hostname,
			ssl_ctx = unitarget.ssl_ctx,
		)
		return target

	
	def __str__(self):
		t = '==== MSLDAPTarget ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t



import os
import zipfile
import json
import base64
import datetime
import asyncio

from tqdm import tqdm

from msldap.external.bloodhoundpy.acls import parse_binary_acl
from msldap.external.bloodhoundpy.resolver import resolve_aces, WELLKNOWN_SIDS
from msldap.external.bloodhoundpy.utils import parse_gplink_string, is_filtered_container, is_filtered_container_child, reverse_dn_components, explode_dn
from msldap.commons.factory import LDAPConnectionFactory
from msldap.connection import MSLDAPClientConnection
from msldap.client import MSLDAPClient
from msldap.commons.adexplorer import Snapshot
from msldap import logger

async def dummy_print(msg):
	print(msg)

class MSLDAPDump2Bloodhound:
	def __init__(self, url: str or MSLDAPClient or LDAPConnectionFactory or MSLDAPClientConnection, progress = True, output_path = None, use_mp:bool=True, print_cb = None):
		self.debug = False
		self.ldap_url = url
		self.connection: MSLDAPClient = None
		self.ldapinfo = None
		self.domainname = None
		self.domainsid = None
		self.use_mp = use_mp
		self.mp_sdbatch_length = 5000
		self.print_cb = print_cb
		self.with_progress = progress
		if self.print_cb is None:
			self.print_cb = dummy_print

		self.DNs = {}
		self.DNs_sorted = {}
		self.ocache = {}
		self.schema = {}
		self.aces = {}
		self.computer_sidcache = {}
		self.token_map = {}

		self.curdate = datetime.datetime.utcnow().strftime('%Y%m%dT%H%M%S')
		self.zipfilepath = '%s_Bloodhound.zip' %  self.curdate
		if output_path is not None:
			self.zipfilepath = os.path.join(output_path, self.zipfilepath)
		self.zipfile = None
		self.MAX_ENTRIES_PER_FILE = 40000
		
		self.totals = {
			'user' : 0,
			'computer' : 0,
			'group' : 0,
			'ou' : 0,
			'gpo' : 0,
			'container' : 0,
			'domain' : 0,
			'trust' : 0
		}
	
	async def print(self, msg:str):
		await self.print_cb(msg)

	
	async def create_progress(self, label, total = None):
		if self.with_progress is True:
			return tqdm(desc = label, total=total)
		else:
			await self.print('[+] %s' % label)
			return None
	
	async def update_progress(self, pbar, value = 1):
		if pbar is None:
			return
		if self.with_progress is True:
			pbar.update(value)
	
	async def close_progress(self, pbar):
		if pbar is None:
			return
		if self.with_progress is True:
			pbar.close()
	
	def get_json_wrapper(self, enumtype):
		return {
			'data' : [],
			'meta': {
				'methods' : 0,
				'type': enumtype,
				'version': 5,
				'count': 0
			}
		}
		

	def split_json(self, enumtype, data):
		if data['meta']['count'] <= self.MAX_ENTRIES_PER_FILE:
			yield data
			return
		
		#split the data
		for i in range(0, data['meta']['count'], self.MAX_ENTRIES_PER_FILE):
			jsonstruct = {
				'data' : [],
				'meta': {
					'methods' : 0,
					'type': enumtype,
					'version': 5,
					'count': 0
				}
			}
			for entry in data['data'][i:i+self.MAX_ENTRIES_PER_FILE]:
				jsonstruct['data'].append(entry)
				jsonstruct['meta']['count'] += 1
			yield jsonstruct

	
	async def write_json_to_zip(self, enumtype, data, filepart = 0):
		if filepart == 0:
			filename = '%s_%s.json' % (self.curdate, enumtype)
		else:
			filename = '%s_%s_%02d.json' % (self.curdate, enumtype, filepart)
		self.zipfile.writestr(filename, json.dumps(data))

	
	async def lookup_dn_children(self, parent_dn):
		parent_dn = parent_dn.upper()
		parent_dn_reversed = reverse_dn_components(parent_dn)
		if parent_dn not in self.DNs:
			logger.debug('[BH] DN not found: %s' % parent_dn_reversed)
			return []

		branch = self.DNs_sorted
		level = 0
		for part in explode_dn(parent_dn_reversed):
			level += 1
			if part not in branch:
				logger.debug('[BH] Part not found: %s Full: %s Branch: %s Level: %s Parts: %s' % (part, parent_dn_reversed, branch.keys(), level, explode_dn(parent_dn_reversed)))
				return []
			branch = branch[part]

		res_dns = []
		for dnpart in branch:
			res_dns.append(dnpart + ',' + parent_dn)
			
		results = []
		for tdn in res_dns:
			if is_filtered_container_child(tdn):
				continue
			if tdn not in self.DNs:
				attrs, err = await self.connection.dnattrs(tdn, ['distinguishedName','objectGUID', 'objectClass','sAMAaccountType', 'sAMAccountName', 'objectSid', 'name'])
				if err is not None:
					raise err
				if attrs is None or len(attrs) == 0:
					logger.debug('[BH] Missing DN: %s' % tdn)
					continue
				res = self.resolve_entry(attrs)
				results.append({
					'ObjectIdentifier': res['objectid'].upper(),
					'ObjectType': res['type'].capitalize(),
				})
				continue
			entry = self.ocache[self.DNs[tdn]]
			results.append({
				'ObjectIdentifier': entry['ObjectIdentifier'].upper(),
				'ObjectType': entry['ObjectType'].capitalize() if entry['ObjectType'].lower() != 'ou' else 'OU',
			})
		
		return results

	async def dump_schema(self):
		pbar = await self.create_progress('Dumping schema')
		# manual stuff here...
		# https://learn.microsoft.com/en-us/windows/win32/adschema/c-foreignsecurityprincipal
		self.schema['foreignsecurityprincipal'] = '89e31c12-8530-11d0-afda-00c04fd930c9'
		
		async for entry, err in self.connection.get_all_schemaentry(['name', 'schemaIDGUID']):
			if err is not None:
				raise err
			await self.update_progress(pbar)
			self.schema[entry.name.lower()] = str(entry.schemaIDGUID)
		await self.close_progress(pbar)

	def add_ocache(self, dn, objectid, principal, otype, dns = '', spns = None):
		self.totals[otype] += 1
		if objectid in WELLKNOWN_SIDS:
			objectid = '%s-%s' % (self.domainname.upper(), objectid.upper())
		self.ocache[objectid] = {
			'dn' : dn.upper(),
			'ObjectIdentifier' : objectid,
			'principal' : principal,
			'ObjectType' : otype,
		}
		self.DNs[dn.upper()] = objectid
		if otype == 'computer':
			entry = {
				'ObjectIdentifier' : objectid,
				'ObjectType' : otype
			}
			if dns is None:
				dns = ''
			self.computer_sidcache[dns.lower()] = entry
			if spns is not None:
				for spn in spns:
					target = spn.split('/')[1]
					target = target.split(':')[0]
					self.computer_sidcache[target.lower()] = entry
	
	def resolve_entry(self, entry):
		# I really REALLY did not want to implement this
		resolved = {}
		account = entry.get('sAMAccountName', '')
		dn = entry.get('distinguishedName', '')
		resolved['objectid'] = entry.get('objectSid', '')
		resolved['principal'] = ('%s@%s' % (account, self.domainname)).upper()
		if 'sAMAaccountName' in entry:
			accountType = entry['sAMAccountType']
			object_class = entry['objectClass']
			if accountType in [268435456, 268435457, 536870912, 536870913]:
				resolved['type'] = 'Group'
			elif accountType in [805306368] or \
				 'msDS-GroupManagedServiceAccount' in object_class or \
				 'msDS-ManagedServiceAccount' in object_class:
				resolved['type'] = 'User'
			elif accountType in [805306369]:
				resolved['type'] = 'Computer'
				short_name = account.rstrip('$')
				resolved['principal'] = ('%s.%s' % (short_name, self.domainname)).upper()
			elif accountType in [805306370]:
				resolved['type'] = 'trustaccount'
			else:
				resolved['type'] = 'Domain'
			return resolved
		
		if 'objectGUID' in entry:
			resolved['objectid'] = entry['objectGUID']
			resolved['principal'] = ('%s@%s' % (entry.get('name', ''), self.domainname)).upper()
			object_class = entry.get('objectClass', [])
			if 'organizationalUnit' in object_class:
				resolved['type'] = 'OU'
			elif 'container' in object_class:
				resolved['type'] = 'Container'
			else:
				resolved['type'] = 'Base'
			return resolved

	async def dump_lookuptable(self):
		pbar = await self.create_progress('Generating lookuptable')
		# domains
		adinfo, err = await self.connection.get_ad_info()
		if err is not None:
			raise err
		self.domainsid = adinfo.objectSid
		self.add_ocache(adinfo.distinguishedName, adinfo.objectSid, '', 'domain')
		await self.update_progress(pbar)

		#trusts
		async for entry, err in self.connection.get_all_trusts(['distinguishedName', 'objectSid', 'objectGUID']):
			if err is not None:
				raise err
			self.add_ocache(entry.distinguishedName, entry.objectGUID, '', 'trust')
			await self.update_progress(pbar)

		#users
		async for entry, err in self.connection.get_all_users(['distinguishedName', 'objectSid', 'objectGUID', 'sAMAccountName']):
			if err is not None:
				raise err
			short_name = entry.sAMAccountName
			self.add_ocache(entry.distinguishedName, entry.objectSid, ('%s@%s' % (short_name, self.domainname)).upper(), 'user')
			await self.update_progress(pbar)

		#machines
		async for entry, err in self.connection.get_all_machines(['distinguishedName', 'objectSid', 'objectGUID', 'sAMAccountName', 'dNSHostName', 'servicePrincipalName']):
			if err is not None:
				raise err
			short_name = entry.sAMAccountName
			dns = entry.dNSHostName
			if dns is None:
				dns = ''

			self.add_ocache(entry.distinguishedName, entry.objectSid, ('%s@%s' % (short_name, self.domainname)).upper(), 'computer', dns, entry.servicePrincipalName)
			await self.update_progress(pbar)

		#groups
		async for entry, err in self.connection.get_all_groups(['distinguishedName', 'objectSid', 'objectGUID']):
			if err is not None:
				raise err
			self.add_ocache(entry.distinguishedName, entry.objectSid, '', 'group')
			await self.update_progress(pbar)

		#ous
		async for entry, err in self.connection.get_all_ous(['distinguishedName', 'objectSid', 'objectGUID']):
			if err is not None:
				raise err
			self.add_ocache(entry.distinguishedName, entry.objectGUID, '', 'ou')
			await self.update_progress(pbar)

		#containers
		async for entry, err in self.connection.get_all_containers(['distinguishedName', 'objectSid', 'objectGUID']):
			if err is not None:
				raise err
			if is_filtered_container(entry.distinguishedName):
				continue
			self.add_ocache(entry.distinguishedName, entry.objectGUID, '', 'container')
			await self.update_progress(pbar)

		#gpos
		async for entry, err in self.connection.get_all_gpos(['distinguishedName', 'objectSid', 'objectGUID']):
			if err is not None:
				raise err
			self.add_ocache(entry.distinguishedName, entry.objectGUID, '', 'gpo')
			await self.update_progress(pbar)

		#foreignsecurityprincipal
		async for entry, err in self.connection.get_all_foreignsecurityprincipals(['name','sAMAccountName', 'objectSid', 'objectGUID', 'distinguishedName', 'objectClass']):
			bhentry = {}
			entry = entry['attributes']
			if 'container' in  entry.get('objectClass', []) is True:
				continue

			if entry['objectSid'] in WELLKNOWN_SIDS:
				bhentry['objectid'] = '%s-%s' % (self.domainname.upper(), entry['objectSid'].upper())
			bhentry['principal'] = self.domainname.upper()
			bhentry['type'] = 'foreignsecurityprincipal'
			if 'name' in entry:
				if entry['name'] in WELLKNOWN_SIDS:
					gname, sidtype = WELLKNOWN_SIDS[entry['name']]
					bhentry['type'] = sidtype.capitalize()
					bhentry['principal'] = '%s@%s' % (gname.upper(), self.domainname.upper())
					bhentry['objectid'] = '%s-%s' % (self.domainname.upper(), entry['objectSid'].upper())
				else:
					bhentry['objectid'] = entry['name']

			self.ocache[bhentry['objectid']] = {
				'dn' : entry['distinguishedName'].upper(),
				'ObjectIdentifier' : bhentry['objectid'],
				'principal' : bhentry['principal'],
				'ObjectType' : bhentry['type'],
			}
			self.DNs[entry['distinguishedName'].upper()] = bhentry['objectid']
			
		await self.close_progress(pbar)

		for dn in [reverse_dn_components(dn) for dn in self.DNs]:
			branch = self.DNs_sorted
			for part in explode_dn(dn):
				if part not in branch:
					branch[part.upper()] = {}
				branch = branch[part.upper()]
		
		if self.debug is True:
			with open('dn.json', 'w') as f:
				json.dump(self.DNs, f, indent=4)

			with open('dntree.json', 'w') as f:
				json.dump(self.DNs_sorted, f, indent=4)
	
	async def dump_acls(self):
		sdbatch = []
		tasks = []
		pbar = await self.create_progress('Dumping SDs', total=len(self.ocache))
		for sid in self.ocache:
			dn = self.ocache[sid]['dn']
			secdesc, err = await self.connection.get_objectacl_by_dn(dn)
			if err is not None:
				raise err
			dn = dn.upper()
			oentry = {
					'IsACLProtected' : None,
					'Properties' : {
						'haslaps' : 'ms-mcs-admpwd' in self.schema
					}
				}
			otype = self.ocache[sid]['ObjectType']
			if otype == 'trust':
				continue
			if otype == 'ou':
				otype = 'organizational-unit'
			if dn.upper() not in self.aces:
				if self.use_mp is True:
					from concurrent.futures import ProcessPoolExecutor
					sdbatch.append((dn, oentry, otype.lower(), secdesc, self.schema))
					if len(sdbatch) > self.mp_sdbatch_length:
						loop = asyncio.get_running_loop()
						with ProcessPoolExecutor() as executor:
							for sde in sdbatch:
								tasks.append(loop.run_in_executor(executor, parse_binary_acl, *sde))
						results = await asyncio.gather(*tasks)
						for dn, aces, relations in results:
							self.aces[dn.upper()] = (aces, relations)
						sdbatch = []
						tasks = []
				else:
					dn, aces, relations = parse_binary_acl(dn, oentry, otype.lower(), secdesc, self.schema)
					self.aces[dn.upper()] = (aces, relations)
			await self.update_progress(pbar)
		
		if len(sdbatch) != 0:
			loop = asyncio.get_running_loop()
			with ProcessPoolExecutor() as executor:
				for sde in sdbatch:
					tasks.append(loop.run_in_executor(executor, parse_binary_acl, *sde))
				results = await asyncio.gather(*tasks)
				for dn, aces, relations in results:
					self.aces[dn.upper()] = (aces, relations)
				sdbatch = []
				tasks = []
		await self.close_progress(pbar)
	
	async def resolve_gplink(self, gplinks):
		if gplinks is None:
			return []

		links = []
		for gplink_dn, options in parse_gplink_string(gplinks):
			link = {}
			link['IsEnforced'] = options == 2
			gplink_dn = gplink_dn.upper()
			if gplink_dn in self.DNs:
				lguid = self.ocache[self.DNs[gplink_dn]]['ObjectIdentifier']
			else:
				attrs, err = await self.connection.dnattrs(gplink_dn.upper(), ['objectGUID', 'objectSid'])
				if err is not None:
					raise err
				if attrs is None or len(attrs) == 0:
					logger.debug('[BH] Missing DN: %s' % gplink_dn)
					continue
				try:
					lguid = attrs['objectGUID']
				except:
					logger.debug('[BH] Missing GUID for %s' % gplink_dn)
					continue
			link['GUID'] = lguid.upper()
			links.append(link)
		return links

	def remove_hidden(self, entry):
		to_del = []
		for k in entry:
			if k.startswith('_'):
				to_del.append(k)
		for k in to_del:
			del entry[k]
		return entry

	async def dump_domains(self):
		pbar = await self.create_progress('Dumping domains', self.totals['domain'])
		adinfo, err = await self.connection.get_ad_info()
		if err is not None:
			raise err
		
		domainentry = adinfo.to_bh(self.domainname)
		
		meta, relations = self.aces[domainentry['Properties']['distinguishedname'].upper()]
		domainentry['IsACLProtected'] = meta['IsACLProtected']
		domainentry['Aces'] = resolve_aces(relations, self.domainname, self.domainsid, self.ocache)
		domainentry['ChildObjects'] =  await self.lookup_dn_children(domainentry['Properties']['distinguishedname'])
		domainentry['Links'] = await self.resolve_gplink(domainentry['_gPLink'])

		jsonstruct = self.get_json_wrapper('domains')
		filectr = 0
		async for entry, err in self.connection.get_all_trusts():
			if err is not None:
				raise err
			domainentry['Trusts'].append(entry.to_bh())
   
		domainentry = self.remove_hidden(domainentry)
		jsonstruct['data'].append(domainentry)
		jsonstruct['meta']['count'] += 1
		if jsonstruct['meta']['count'] == self.MAX_ENTRIES_PER_FILE:
			await self.write_json_to_zip('domains', jsonstruct, filectr)
			jsonstruct = self.get_json_wrapper('domains')
			filectr += 1
		await self.update_progress(pbar)
		
		if jsonstruct['meta']['count'] > 0:
			await self.write_json_to_zip('domains', jsonstruct, filectr)
		await self.close_progress(pbar)
		if self.debug is True:
			with open('domains.json', 'w') as f:
				json.dump(jsonstruct, f)
	
	async def dump_users(self):
		pbar = await self.create_progress('Dumping users', self.totals['user'])

		jsonstruct = self.get_json_wrapper('users')
		filectr = 0
		async for ldapentry, err in self.connection.get_all_users():
			if err is not None:
				raise err
			
			entry = ldapentry.to_bh(self.domainname)
			meta, relations = self.aces[entry['Properties']['distinguishedname'].upper()]
			entry['IsACLProtected'] = meta['IsACLProtected']
			entry['Aces'] = resolve_aces(relations, self.domainname, self.domainsid, self.ocache)
			
			if entry['_allowerdtodelegateto'] is not None:
				seen = {}
				for host in entry['_allowerdtodelegateto']:
					try:
						target = host.split('/')[1]
						target = target.split(':')[0]
					except IndexError:
						logger.debug('[BH] Invalid delegation target: %s', host)
						continue
					try:
						sid = self.computer_sidcache[target.lower()]
						if sid['ObjectIdentifier'] in seen:
							continue
						seen[sid['ObjectIdentifier']] = 1
						entry['AllowedToDelegate'].append(sid)
					except KeyError:
						if '.' in target:
							entry['AllowedToDelegate'].append(target.upper())
			entry = self.remove_hidden(entry)

			jsonstruct['data'].append(entry)
			jsonstruct['meta']['count'] += 1
			if jsonstruct['meta']['count'] == self.MAX_ENTRIES_PER_FILE:
				await self.write_json_to_zip('users', jsonstruct, filectr)
				jsonstruct = self.get_json_wrapper('users')
				filectr += 1
			await self.update_progress(pbar)
		
		if jsonstruct['meta']['count'] > 0:
			await self.write_json_to_zip('users', jsonstruct, filectr)
		await self.close_progress(pbar)

		if self.debug is True:
			with open('users.json', 'w') as f:
				json.dump(jsonstruct, f)
	
	async def dump_computers(self):
		pbar = await self.create_progress('Dumping computers', self.totals['computer'])
		jsonstruct = self.get_json_wrapper('computers')
		filectr = 0
		async for ldapentry, err in self.connection.get_all_machines():
			entry = ldapentry.to_bh(self.domainname)
			meta, relations = self.aces[entry['Properties']['distinguishedname'].upper()]
			entry['IsACLProtected'] = meta['IsACLProtected']
			entry['Aces'] = resolve_aces(relations, self.domainname, self.domainsid, self.ocache)
			
			if entry['_allowedtoactonbehalfofotheridentity'] is not None:
				allowedacl = base64.b64decode(entry['_allowedtoactonbehalfofotheridentity'])
				_, entryres, relations = parse_binary_acl(entry['Properties']['distinguishedname'].upper(), entry, 'computer', allowedacl, self.schema)
				
				for ace in resolve_aces(relations, self.domainname, self.domainsid, self.ocache):
					if ace['RightName'] == 'Owner':
						continue
					if ace['RightName'] == 'GenericAll':
						entryres['AllowedToAct'].append({
							'ObjectIdentifier': ace['PrincipalSID'], 
							'ObjectType': ace['PrincipalType'].capitalize()
						})
			
			del entry['_allowedtoactonbehalfofotheridentity']
			if entry['Properties']['allowedtodelegate'] is not None:
				seen = {}
				for host in entry['Properties']['allowedtodelegate']:
					try:
						target = host.split('/')[1]
						target = target.split(':')[0]
					except IndexError:
						logger.debug('[BH] Invalid delegation target: %s', host)
						continue
					try:
						sid = self.computer_sidcache[target.lower()]
						if sid['ObjectIdentifier'] in seen:
							continue
						seen[sid['ObjectIdentifier']] = 1
						entry['AllowedToDelegate'].append(sid)
					except KeyError:
						if '.' in target:
							entry['AllowedToDelegate'].append({
								"ObjectIdentifier": target.upper(),
								"ObjectType": "Computer"
							})

			entry = self.remove_hidden(entry)
			jsonstruct['data'].append(entry)
			jsonstruct['meta']['count'] += 1
			if jsonstruct['meta']['count'] == self.MAX_ENTRIES_PER_FILE:
				await self.write_json_to_zip('computers', jsonstruct, filectr)
				jsonstruct = self.get_json_wrapper('computers')
				filectr += 1
			await self.update_progress(pbar)
		
		if jsonstruct['meta']['count'] > 0:
			await self.write_json_to_zip('computers', jsonstruct, filectr)
		await self.close_progress(pbar)

		if self.debug is True:
			with open('computers.json', 'w') as f:
				json.dump(jsonstruct, f)
   
	async def dump_groups(self):
		pbar = await self.create_progress('Dumping groups', self.totals['group'])
		jsonstruct = self.get_json_wrapper('groups')
		filectr = 0
		async for ldapentry, err in self.connection.get_all_groups():
			entry = ldapentry.to_bh(self.domainname)
			meta, relations = self.aces[entry['Properties']['distinguishedname'].upper()]
			entry['IsACLProtected'] = meta['IsACLProtected']
			entry['Aces'] = resolve_aces(relations, self.domainname, self.domainsid, self.ocache)
			
			if ldapentry.member is not None:
				for member in ldapentry.member:
					if member.upper() in self.DNs:
						oid = self.DNs[member.upper()]
						entry['Members'].append({
							'ObjectIdentifier' : self.ocache[oid]['ObjectIdentifier'],
							'ObjectType' : self.ocache[oid]['ObjectType'].capitalize()
						})
					else:
						if member.find('ForeignSecurityPrincipals') != -1:
							continue
	  
			entry = self.remove_hidden(entry)
			jsonstruct['data'].append(entry)
			jsonstruct['meta']['count'] += 1
			if jsonstruct['meta']['count'] == self.MAX_ENTRIES_PER_FILE:
				await self.write_json_to_zip('groups', jsonstruct, filectr)
				jsonstruct = self.get_json_wrapper('groups')
				filectr += 1
			await self.update_progress(pbar)
		
		if jsonstruct['meta']['count'] > 0:
			await self.write_json_to_zip('groups', jsonstruct, filectr)
		await self.close_progress(pbar)

		if self.debug is True:
			with open('groups.json', 'w') as f:
				json.dump(jsonstruct, f)

	async def dump_gpos(self):
		pbar = await self.create_progress('Dumping GPOs', self.totals['gpo'])
		jsonstruct = self.get_json_wrapper('gpos')
		filectr = 0
		async for ldapentry, err in self.connection.get_all_gpos():
			entry = ldapentry.to_bh(self.domainname, self.domainsid)
			meta, relations = self.aces[entry['Properties']['distinguishedname'].upper()]
			entry['IsACLProtected'] = meta['IsACLProtected']
			entry['Aces'] = resolve_aces(relations, self.domainname, self.domainsid, self.ocache)      
			entry = self.remove_hidden(entry)

			jsonstruct['data'].append(entry)
			jsonstruct['meta']['count'] += 1
			if jsonstruct['meta']['count'] == self.MAX_ENTRIES_PER_FILE:
				await self.write_json_to_zip('gpos', jsonstruct, filectr)
				jsonstruct = self.get_json_wrapper('gpos')
				filectr += 1
			await self.update_progress(pbar)
		
		if jsonstruct['meta']['count'] > 0:
			await self.write_json_to_zip('gpos', jsonstruct, filectr)
		await self.close_progress(pbar)

		if self.debug is True:
			with open('gpos.json', 'w') as f:
				json.dump(jsonstruct, f)

	async def dump_ous(self):
		pbar = await self.create_progress('Dumping OUs', self.totals['ou'])
		jsonstruct = self.get_json_wrapper('ous')
		filectr = 0

		async for ldapentry, err in self.connection.get_all_ous():
			if err is not None:
				raise err
			entry = ldapentry.to_bh(self.domainname, self.domainsid)
			meta, relations = self.aces[entry['Properties']['distinguishedname'].upper()]
			entry['IsACLProtected'] = meta['IsACLProtected']
			entry['Aces'] = resolve_aces(relations, self.domainname, self.domainsid, self.ocache)
			entry['ChildObjects'] =  await self.lookup_dn_children(entry['Properties']['distinguishedname'])
			entry['Links'] = await self.resolve_gplink(entry['_gPLink'])
			entry = self.remove_hidden(entry)

			jsonstruct['data'].append(entry)
			jsonstruct['meta']['count'] += 1
			if jsonstruct['meta']['count'] == self.MAX_ENTRIES_PER_FILE:
				await self.write_json_to_zip('ous', jsonstruct, filectr)
				jsonstruct = self.get_json_wrapper('ous')
				filectr += 1

			await self.update_progress(pbar)
		
		if jsonstruct['meta']['count'] > 0:
			await self.write_json_to_zip('ous', jsonstruct, filectr)
		await self.close_progress(pbar)

		if self.debug is True:
			with open('ous.json', 'w') as f:
				json.dump(jsonstruct, f)
	
	async def dump_containers(self):
		pbar = await self.create_progress('Dumping Containers', self.totals['container'])
		jsonstruct = self.get_json_wrapper('containers')
		filectr = 0
		async for ldapentry, err in self.connection.get_all_containers():
			if err is not None:
				raise err
			if is_filtered_container(ldapentry.distinguishedName):
				continue
			entry = ldapentry.to_bh(self.domainname, self.domainsid)
			meta, relations = self.aces[entry['Properties']['distinguishedname'].upper()]
			entry['IsACLProtected'] = meta['IsACLProtected']
			entry['Aces'] = resolve_aces(relations, self.domainname, self.domainsid, self.ocache)
			entry['ChildObjects'] =  await self.lookup_dn_children(entry['Properties']['distinguishedname'])
			entry = self.remove_hidden(entry)

			jsonstruct['data'].append(entry)
			jsonstruct['meta']['count'] += 1
			if jsonstruct['meta']['count'] == self.MAX_ENTRIES_PER_FILE:
				await self.write_json_to_zip('containers', jsonstruct, filectr)
				jsonstruct = self.get_json_wrapper('containers')
				filectr += 1
			await self.update_progress(pbar)
		
		if jsonstruct['meta']['count'] > 0:
			await self.write_json_to_zip('containers', jsonstruct, filectr)
		await self.close_progress(pbar)

		if self.debug is True:
			with open('containers.json', 'w') as f:
				json.dump(jsonstruct, f)
	
	async def dump_ldap(self):
		if isinstance(self.ldap_url, str):
			if self.ldap_url.startswith('adexplorer://'):
				self.ldap_url = self.ldap_url[13:]
				await self.print('[+] Parsing ADEXPLORER Snapshot...')
				self.connection = await Snapshot.from_file(self.ldap_url)
				self.ldap_url = self.connection
				await self.print('[+] Parsing done!')

		if isinstance(self.ldap_url, Snapshot) is False:
			if isinstance(self.ldap_url, str):
				factory = LDAPConnectionFactory.from_url(self.ldap_url)
				self.connection = factory.get_client()
				self.connection.keepalive = True
			if isinstance(self.ldap_url, LDAPConnectionFactory):
				self.connection = self.ldap_url.get_client()
				self.connection.keepalive = True
			if isinstance(self.ldap_url, MSLDAPClient):
				self.connection = self.ldap_url
			
			if isinstance(self.ldap_url, MSLDAPClientConnection):
				self.connection = MSLDAPClient(None, None, connection = self.ldap_url)
			
			await self.print('[+] Connecting to LDAP server')
			self.connection.keepalive = True
			_, err = await self.connection.connect()
			if err is not None:
				raise err
			await self.print('[+] Connected to LDAP serrver')
			
			self.ldapinfo = self.connection.get_server_info()
			self.domainname = self.ldapinfo['defaultNamingContext'].upper().replace('DC=','').replace(',','.')
		else:
			self.domainname = self.connection.rootdomain.upper().replace('DC=','').replace(',','.')
		
		await self.dump_schema()
		await self.dump_lookuptable()
		await self.dump_acls()
		with zipfile.ZipFile(self.zipfilepath, 'w', zipfile.ZIP_DEFLATED) as self.zipfile:
			await self.dump_domains()
			await self.dump_users()
			await self.dump_computers()
			await self.dump_groups()
			await self.dump_gpos()
			await self.dump_ous()
			await self.dump_containers()
		await self.print('[+] Bloodhound data saved to %s' % self.zipfilepath)

	
	async def run(self):
		await self.dump_ldap()